#include <stdio.h>
#include <stdarg.h>
#if defined( sgi )
#include <alloca.h>
#else
#include <stdlib.h>
#endif

#include "mpi.h"


/*@

  RGX_List_Processor_Names - List all the processor/machine names in the
                             given Communicator either in stdout or in
                             returned character array.  There are 2
                             different calling interfaces which are listed
                             as follows,

.vb
  int RGX_List_Processor_Names( MPI_Comm  World_Comm,
                                int N_task_out = -1 )

  int RGX_List_Processor_Names( MPI_Comm  World_Comm,
                                int      *N_task_out_ptr,
                                char     *processor_names_out )
.ve

  Unmodified Input Variables :
. World_Comm - Input Communicator containing all processors for the grid
               (Usually MPI_COMM_WORLD)
. N_task_out_ptr - Pointer to Number of Tasks, N_task, in the Communicator.
                   If N_task = -1, pointer to the character array of the
                   processor names is NOT assumed.
                   If N_task > 0, a pointer to a character array of 
                   sufficient size is needed.  N_task inputed is the number 
                   of the processors names the array can hold.  If N_task <
                   number of processors in the communicator, only the first
                   N_task of the processor names are returned.

  Modified Output Variables :
. processor_names_out - Pointer to the Character Array, of size at least
                        bigger than Number of processors times 
                        MPI_MAX_PROCESSOR_NAME.  The array will contain
                        all the processor names if N_task > 0.  Where
                        ii th processor name can be accessed at
                        processor_names_out[ ii * MPI_MAX_PROCESSOR_NAME ].
. returned value - returns MPI_SUCCESS if completed successfully
.n

.N RGX_fortran
.N RGX_minimal_set_of_variables
.N RGX_supported_datatype
.N RGX_yes_included

@*/
int RGX_List_Processor_Names(
                              MPI_Comm  World_Comm,

                              ...
                            )
{
#   define       line_Max  MPI_MAX_PROCESSOR_NAME

/*  Variable argument list  */
    va_list      ap;
    int         *N_task_out_ptr;
    char        *processor_names_out;

/*  Local variables  */
    int          N_task;
    char        *processor_names;
    int         *proc_name_lgths;
    int         *displacements;
    int         *Ones;
    int         *Ones_displ;
    const int    master_tid = 0;
    int          my_tid;
    char         local_node_name[ line_Max ];
    int          local_name_lgth;
    int          ierr;
    int          ii;


     /*  Find the No. of Tasks involved in the World_Comm  */

    ierr = MPI_Comm_rank( World_Comm, &my_tid );
    ierr = MPI_Comm_size( World_Comm, &N_task );
    ierr = MPI_Get_processor_name( local_node_name, &local_name_lgth );

    /*  Increment local_name_lgth to account for the possible missing '\0'  */
    local_name_lgth += 1;

    if ( my_tid == master_tid )
       fprintf( stdout, "There are %d tasks running\n", N_task );

    if ( my_tid == master_tid )
    {
       processor_names = ( char* ) alloca( N_task * line_Max * sizeof( char ) );
       proc_name_lgths = ( int* )  alloca( N_task * sizeof( int ) );
       displacements   = ( int* )  alloca( N_task * sizeof( int ) );
       Ones            = ( int* )  alloca( N_task * sizeof( int ) );
       Ones_displ      = ( int* )  alloca( N_task * sizeof( int ) );
       for ( ii = 0; ii < N_task; ii++ )
       {
          displacements[ ii ] = ii * line_Max;
          Ones[ ii ]          = 1;
          Ones_displ[ ii ]    = ii;
       }
    }

    /*  Increment the local_name_lgth to account for '\0'  */
    /*  local_name_lgth += 1;                              */

    ierr = MPI_Gatherv( &local_name_lgth, 1, MPI_INT,
                        proc_name_lgths, Ones, Ones_displ,
                        MPI_INT, master_tid, World_Comm );
    if ( ierr != MPI_SUCCESS )
    {
       fprintf( stderr, __FILE__": MPI_Gatherv( &local_name_lgth ) fails "
                        "with ierr = %d\n", ierr );
       return ierr;
    }

    ierr = MPI_Gatherv( local_node_name, local_name_lgth, MPI_CHAR,
                        processor_names, proc_name_lgths, displacements,
                        MPI_CHAR, master_tid, World_Comm );
    if ( ierr != MPI_SUCCESS )
    {
       fprintf( stderr, __FILE__": MPI_Gatherv( local_node_name ) fails "
                        "with ierr = %d\n", ierr );
       return ierr;
    }

    va_start( ap, World_Comm );
    N_task_out_ptr = va_arg( ap, int* );

    if ( *N_task_out_ptr > 0 )
    {
       /* Initialize pointer, processor_names_out for Master and Slaves.  */
       processor_names_out = va_arg( ap, char* );

       /*  Use of processor_names_out to main memory consistency  */
       /*  So the local allocated memory, processor_names, can be freed  */
       if ( my_tid == master_tid )
       {
          if ( *N_task_out_ptr >= N_task )
          {
             *N_task_out_ptr = N_task;
             for ( ii = 0; ii < *N_task_out_ptr; ii++ )
                strcpy( &processor_names_out[ ii*line_Max ],
                        &processor_names[ ii*line_Max ] );
          }
          else
          {
             for ( ii = 0; ii < *N_task_out_ptr; ii++ )
                strcpy( &processor_names_out[ ii*line_Max ],
                        &processor_names[ ii*line_Max ] );
             fprintf( stderr, __FILE__": Input array size is too small,"
                              "N_task = %d > N_task_out = %d\n",
                               N_task, *N_task_out_ptr );
          }
       }

       /*  Broadcast N_task and processor_names[] back to all slaves  */
       ierr = MPI_Bcast( N_task_out_ptr, 1, MPI_INT, master_tid, World_Comm );
       ierr = MPI_Bcast( processor_names_out, 
                         ( *N_task_out_ptr ) * line_Max,
                         MPI_CHAR, master_tid, World_Comm );
       /*
       printf( "tid=%d:  The *N_task_out_ptr = %d\n",
               my_tid, *N_task_out_ptr );
       for ( ii = 0; ii < *N_task_out_ptr; ii++ )
          printf( "tid=%d: processor_names_out[ %d ] = %s\n",
                  my_tid, ii, &processor_names_out[ii*line_Max] );
       */
    }

    va_end( ap );
                      

    /*  Print the list of machines in Master's StdOut and then cleanup  */
    if ( my_tid == master_tid )
    {
       for ( ii = 0; ii < N_task; ii++ )
       {
          local_name_lgth = proc_name_lgths[ ii ];
          strncpy( local_node_name, &processor_names[ ii*line_Max ],
                   local_name_lgth );
          fprintf( stdout, "%d, One task lives on %s\n",
                           ii, local_node_name );
       }
       printf( "\n" );
    }

    return MPI_SUCCESS;
}
